from __future__ import annotations

import typing
import logging
import datetime
from redis.commands import (
    CoreCommands,
    AsyncCoreCommands,
    RedisModuleCommands,
    SentinelCommands,
    AsyncSentinelCommands,
)
from redis.client import (
    #AbstractRedis,
    Redis,
    Pipeline,
    PubSub,
    CaseInsensitiveDict,
    parse_scan,
    
)
from redis.asyncio.client import (
    Redis as AsyncRedis,
    PubSub as AsyncPubSub,
    Pipeline as AsyncPipeline,
)

from aiokeydb.v2.connection import (
    ConnectionPool,
    AsyncConnectionPool,
)

from typing import Any, Iterable, Mapping, Callable, Union, overload, TYPE_CHECKING
from typing_extensions import Literal

if TYPE_CHECKING:
    from redis.typing import ChannelT
    from redis.client import _Key, _Value, _StrType

logger = logging.getLogger(__name__)


"""
Retryable Components for AsyncKeyDB
"""
from .utils.helpers import get_retryable_wrapper

retryable_wrapper = get_retryable_wrapper()


class RetryablePubSub(PubSub):
    """
    Retryable PubSub
    """

    @retryable_wrapper
    def subscribe(self, *args: 'ChannelT', **kwargs: typing.Callable):
        """
        Subscribe to channels. Channels supplied as keyword arguments expect
        a channel name as the key and a callable as the value. A channel's
        callable will be invoked automatically when a message is received on
        that channel rather than producing a message via ``listen()`` or
        ``get_message()``.
        """
        return super().subscribe(*args, **kwargs)
    
    @retryable_wrapper
    def unsubscribe(self, *args):
        """
        Unsubscribe from the supplied channels. If empty, unsubscribe from
        all channels
        """
        return super().unsubscribe(*args)

    @retryable_wrapper
    def listen(self) -> typing.Iterator:
        """Listen for messages on channels this client has been subscribed to"""
        yield from super().listen()


class RetryablePipeline(Pipeline):
    """
    Retryable Pipeline
    """

    @retryable_wrapper
    def execute(self, raise_on_error: bool = True) -> typing.List[typing.Any]:
        """Execute all the commands in the current pipeline"""
        return super().execute(raise_on_error = raise_on_error)


PubSubT = Union[PubSub, RetryablePubSub]
PipelineT = Union[Pipeline, RetryablePipeline]

class KeyDB(Redis):
    """
    Implementation of the KeyDB protocol.

    This abstract class provides a Python interface to all KeyDB commands
    and an implementation of the KeyDB protocol.

    Pipelines derive from this, implementing how
    the commands are sent and received to the KeyDB server. Based on
    configuration, an instance will either use a ConnectionPool, or
    Connection object to talk to keydb.

    It is not safe to pass PubSub or Pipeline objects between threads.
    """

    @property
    def is_async(self):
        return False

    @classmethod
    def from_url(
        cls, 
        url, 
        pool_class: typing.Optional[typing.Type[ConnectionPool]] = ConnectionPool,
        **kwargs
    ):
        """
        Return a Redis client object configured from the given URL

        For example::

            keydb://[[username]:[password]]@localhost:6379/0
            keydbs://[[username]:[password]]@localhost:6379/0
            redis://[[username]:[password]]@localhost:6379/0
            rediss://[[username]:[password]]@localhost:6379/0
            unix://[username@]/path/to/socket.sock?db=0[&password=password]

        Five URL schemes are supported:

        - `keydb://` creates a TCP socket connection.
        - `keydbs://` creates a SSL wrapped TCP socket connection.
        - `redis://` creates a TCP socket connection. See more at:
          <https://www.iana.org/assignments/uri-schemes/prov/redis>
        - `rediss://` creates a SSL wrapped TCP socket connection. See more at:
          <https://www.iana.org/assignments/uri-schemes/prov/rediss>
        - ``unix://``: creates a Unix Domain Socket connection.

        The username, password, hostname, path and all querystring values
        are passed through urllib.parse.unquote in order to replace any
        percent-encoded values with their corresponding characters.

        There are several ways to specify a database number. The first value
        found will be used:

            1. A ``db`` querystring option, e.g. redis://localhost?db=0
            2. If using the redis:// or rediss:// schemes, the path argument
               of the url, e.g. redis://localhost/0
            3. A ``db`` keyword argument to this function.

        If none of these options are specified, the default db=0 is used.

        All querystring options are cast to their appropriate Python types.
        Boolean arguments can be specified with string values "True"/"False"
        or "Yes"/"No". Values that cannot be properly cast cause a
        ``ValueError`` to be raised. Once parsed, the querystring arguments
        and keyword arguments are passed to the ``ConnectionPool``'s
        class initializer. In the case of conflicting arguments, querystring
        arguments always win.

        """
        if not pool_class: pool_class = ConnectionPool
        connection_pool = pool_class.from_url(url, **kwargs)
        return cls(connection_pool=connection_pool)


    def pubsub(
        self, 
        retryable: typing.Optional[bool] = False,
        **kwargs,
    ) -> PubSubT:
        """
        Return a Publish/Subscribe object. With this object, you can
        subscribe to channels and listen for messages that get published to
        them.
        """
        pubsub_class = RetryablePubSub if retryable else PubSub
        return pubsub_class(connection_pool = self.connection_pool, **kwargs)
    

    def pipeline(
        self, 
        transaction: bool = True, 
        shard_hint: typing.Optional[str] = None,
        retryable: typing.Optional[bool] = False,
    ) -> PipelineT:
        """
        Return a new pipeline object that can queue multiple commands for
        later execution. ``transaction`` indicates whether all commands
        should be executed atomically. Apart from making a group of operations
        atomic, pipelines are useful for reducing the back-and-forth overhead
        between the client and server.
        """
        pipeline_class = RetryablePipeline if retryable else Pipeline
        return pipeline_class(
            connection_pool = self.connection_pool, 
            response_callbacks = self.response_callbacks, 
            transaction = transaction, 
            shard_hint = shard_hint
        )



class AsyncRetryablePubSub(AsyncPubSub):
    """
    Retryable PubSub
    """

    @retryable_wrapper
    async def subscribe(self, *args: 'ChannelT', **kwargs: typing.Callable):
        """
        Subscribe to channels. Channels supplied as keyword arguments expect
        a channel name as the key and a callable as the value. A channel's
        callable will be invoked automatically when a message is received on
        that channel rather than producing a message via ``listen()`` or
        ``get_message()``.
        """
        return await super().subscribe(*args, **kwargs)
    
    @retryable_wrapper
    def unsubscribe(self, *args) -> typing.Awaitable:
        """
        Unsubscribe from the supplied channels. If empty, unsubscribe from
        all channels
        """
        return super().unsubscribe(*args)

    @retryable_wrapper
    async def listen(self) -> typing.AsyncIterator:
        """Listen for messages on channels this client has been subscribed to"""
        async for response in super().listen():
            yield response

class AsyncRetryablePipeline(AsyncPipeline):
    """
    Retryable Pipeline
    """

    @retryable_wrapper
    async def execute(self, raise_on_error: bool = True):
        """Execute all the commands in the current pipeline"""
        return await super().execute(raise_on_error = raise_on_error)
    
    if typing.TYPE_CHECKING:
        # This is only for type checking purposes, and will be removed in the future.
        def exists(self, *names: _Key) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def expire(  # type: ignore[override]
            self, name: _Key, time: int | datetime.timedelta, nx: bool = False, xx: bool = False, gt: bool = False, lt: bool = False
        ) -> 'AsyncRetryablePipeline': ...
        def expireat(self, name, when, nx: bool = False, xx: bool = False, gt: bool = False, lt: bool = False) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def get(self, name: _Key) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def getdel(self, name: _Key) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def getex(  # type: ignore[override]
            self,
            name,
            ex: Any | None = None,
            px: Any | None = None,
            exat: Any | None = None,
            pxat: Any | None = None,
            persist: bool = False,
        ) -> 'AsyncRetryablePipeline': ...
        def set(  # type: ignore[override]
            self,
            name: _Key,
            value: _Value,
            ex: None | int | datetime.timedelta = None,
            px: None | int | datetime.timedelta = None,
            nx: bool = False,
            xx: bool = False,
            keepttl: bool = False,
            get: bool = False,
            exat: Any | None = None,
            pxat: Any | None = None,
        ) -> 'AsyncRetryablePipeline': ...
        def setbit(self, name: _Key, offset: int, value: int) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def setex(self, name: _Key, time: int | datetime.timedelta, value: _Value) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def setnx(self, name: _Key, value: _Value) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def setrange(self, name, offset, value) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def stralgo(  # type: ignore[override]
            self,
            algo,
            value1,
            value2,
            specific_argument: str = "strings",
            len: bool = False,
            idx: bool = False,
            minmatchlen: Any | None = None,
            withmatchlen: bool = False,
            **kwargs: Any,
        ) -> 'AsyncRetryablePipeline': ...
        def strlen(self, name) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def substr(self, name, start, end: int = -1) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def touch(self, *args) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def ttl(self, name: _Key) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        @overload
        def blpop(self, keys: _Value | Iterable[_Value], timeout: Literal[0] | None = 0) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        @overload
        def blpop(self, keys: _Value | Iterable[_Value], timeout: float) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        @overload
        def brpop(self, keys: _Value | Iterable[_Value], timeout: Literal[0] | None = 0) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        @overload
        def brpop(self, keys: _Value | Iterable[_Value], timeout: float) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def brpoplpush(self, src, dst, timeout: int | None = 0) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def lindex(self, name: _Key, index: int) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def linsert(  # type: ignore[override]
            self, name: _Key, where: Literal["BEFORE", "AFTER", "before", "after"], refvalue: _Value, value: _Value
        ) -> 'AsyncRetryablePipeline': ...
        def llen(self, name: _Key) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def lpop(self, name, count: int | None = None) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def lpush(self, name: _Value, *values: _Value) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def lpushx(self, name, value) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def lrange(self, name: _Key, start: int, end: int) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def lrem(self, name: _Key, count: int, value: _Value) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def lset(self, name: _Key, index: int, value: _Value) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def ltrim(self, name: _Key, start: int, end: int) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def rpop(self, name, count: int | None = None) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def rpoplpush(self, src, dst) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def rpush(self, name: _Value, *values: _Value) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def rpushx(self, name, value) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]

        def zadd(  # type: ignore[override]
        self,
        name: _Key,
        mapping: Mapping[_Key, _Value],
        nx: bool = False,
        xx: bool = False,
        ch: bool = False,
        incr: bool = False,
        gt: Any | None = False,
        lt: Any | None = False,
        ) -> 'AsyncRetryablePipeline': ...
        def zcard(self, name: _Key) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def zcount(self, name: _Key, min: _Value, max: _Value) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def zdiff(self, keys, withscores: bool = False) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def zdiffstore(self, dest, keys) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def zincrby(self, name: _Key, amount: float, value: _Value) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def zinter(self, keys, aggregate: Any | None = None, withscores: bool = False) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def zinterstore(self, dest: _Key, keys: Iterable[_Key], aggregate: Literal["SUM", "MIN", "MAX"] | None = None) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def zlexcount(self, name: _Key, min: _Value, max: _Value) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def zpopmax(self, name: _Key, count: int | None = None) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def zpopmin(self, name: _Key, count: int | None = None) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def zrandmember(self, key, count: Any | None = None, withscores: bool = False) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        @overload  # type: ignore[override]
        def bzpopmax(self, keys: _Key | Iterable[_Key], timeout: Literal[0] = 0) -> 'AsyncRetryablePipeline': ...
        @overload  # type: ignore[override]
        def bzpopmax(self, keys: _Key | Iterable[_Key], timeout: float) -> 'AsyncRetryablePipeline': ...
        @overload  # type: ignore[override]
        def bzpopmin(self, keys: _Key | Iterable[_Key], timeout: Literal[0] = 0) -> 'AsyncRetryablePipeline': ...
        @overload  # type: ignore[override]
        def bzpopmin(self, keys: _Key | Iterable[_Key], timeout: float) -> 'AsyncRetryablePipeline': ...
        @overload  # type: ignore[override]
        def zrange(
            self,
            name: _Key,
            start: int,
            end: int,
            desc: bool,
            withscores: Literal[True],
            score_cast_func: Callable[[_StrType], Any],
            byscore: bool = False,
            bylex: bool = False,
            offset: int | None = None,
            num: int | None = None,
        ) -> 'AsyncRetryablePipeline': ...
        @overload  # type: ignore[override]
        def zrange(
            self,
            name: _Key,
            start: int,
            end: int,
            desc: bool,
            withscores: Literal[True],
            score_cast_func: Callable[[_StrType], float] = ...,
            byscore: bool = False,
            bylex: bool = False,
            offset: int | None = None,
            num: int | None = None,
        ) -> 'AsyncRetryablePipeline': ...
        @overload  # type: ignore[override]
        def zrange(
            self,
            name: _Key,
            start: int,
            end: int,
            *,
            withscores: Literal[True],
            score_cast_func: Callable[[_StrType], None],
            byscore: bool = False,
            bylex: bool = False,
            offset: int | None = None,
            num: int | None = None,
        ) -> 'AsyncRetryablePipeline': ...
        @overload  # type: ignore[override]
        def zrange(
            self,
            name: _Key,
            start: int,
            end: int,
            *,
            withscores: Literal[True],
            score_cast_func: Callable[[_StrType], float] = ...,
            byscore: bool = False,
            bylex: bool = False,
            offset: int | None = None,
            num: int | None = None,
        ) -> 'AsyncRetryablePipeline': ...
        @overload  # type: ignore[override]
        def zrange(
            self,
            name: _Key,
            start: int,
            end: int,
            desc: bool = False,
            withscores: bool = False,
            score_cast_func: Callable[[_StrType], Any] = ...,
            byscore: bool = False,
            bylex: bool = False,
            offset: int | None = None,
            num: int | None = None,
        ) -> 'AsyncRetryablePipeline': ...
        @overload  # type: ignore[override]
        def zrevrange(
            self, name: _Key, start: int, end: int, withscores: Literal[True], score_cast_func: Callable[[_StrType], None]
        ) -> 'AsyncRetryablePipeline': ...
        @overload  # type: ignore[override]
        def zrevrange(self, name: _Key, start: int, end: int, withscores: Literal[True]) -> 'AsyncRetryablePipeline': ...
        @overload  # type: ignore[override]
        def zrevrange(
            self, name: _Key, start: int, end: int, withscores: bool = False, score_cast_func: Callable[[Any], Any] = ...
        ) -> 'AsyncRetryablePipeline': ...
        def zrangestore(  # type: ignore[override]
            self,
            dest,
            name,
            start,
            end,
            byscore: bool = False,
            bylex: bool = False,
            desc: bool = False,
            offset: Any | None = None,
            num: Any | None = None,
        ) -> 'AsyncRetryablePipeline': ...
        def zrangebylex(self, name: _Key, min: _Value, max: _Value, start: int | None = None, num: int | None = None) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def zrevrangebylex(self, name: _Key, max: _Value, min: _Value, start: int | None = None, num: int | None = None) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        @overload  # type: ignore[override]
        def zrangebyscore(
            self,
            name: _Key,
            min: _Value,
            max: _Value,
            start: int | None = None,
            num: int | None = None,
            *,
            withscores: Literal[True],
            score_cast_func: Callable[[_StrType], None],
        ) -> 'AsyncRetryablePipeline': ...
        @overload  # type: ignore[override]
        def zrangebyscore(
            self, name: _Key, min: _Value, max: _Value, start: int | None = None, num: int | None = None, *, withscores: Literal[True]
        ) -> 'AsyncRetryablePipeline': ...
        @overload  # type: ignore[override]
        def zrangebyscore(
            self,
            name: _Key,
            min: _Value,
            max: _Value,
            start: int | None = None,
            num: int | None = None,
            withscores: bool = False,
            score_cast_func: Callable[[_StrType], Any] = ...,
        ) -> 'AsyncRetryablePipeline': ...
        @overload  # type: ignore[override]
        def zrevrangebyscore(
            self,
            name: _Key,
            max: _Value,
            min: _Value,
            start: int | None = None,
            num: int | None = None,
            *,
            withscores: Literal[True],
            score_cast_func: Callable[[_StrType], Any],
        ) -> 'AsyncRetryablePipeline': ...
        @overload  # type: ignore[override]
        def zrevrangebyscore(
            self, name: _Key, max: _Value, min: _Value, start: int | None = None, num: int | None = None, *, withscores: Literal[True]
        ) -> 'AsyncRetryablePipeline': ...
        @overload  # type: ignore[override]
        def zrevrangebyscore(
            self,
            name: _Key,
            max: _Value,
            min: _Value,
            start: int | None = None,
            num: int | None = None,
            withscores: bool = False,
            score_cast_func: Callable[[_StrType], Any] = ...,
        ) -> 'AsyncRetryablePipeline': ...
        def zrank(self, name: _Key, value: _Value) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def zrem(self, name: _Key, *values: _Value) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def zremrangebylex(self, name: _Key, min: _Value, max: _Value) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def zremrangebyrank(self, name: _Key, min: int, max: int) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def zremrangebyscore(self, name: _Key, min: _Value, max: _Value) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def zrevrank(self, name: _Key, value: _Value, withscore: bool = False) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def zscore(self, name: _Key, value: _Value) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def zunion(self, keys, aggregate: Any | None = None, withscores: bool = False) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def zunionstore(self, dest: _Key, keys: Iterable[_Key], aggregate: Literal["SUM", "MIN", "MAX"] | None = None) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]
        def zmscore(self, key, members) -> 'AsyncRetryablePipeline': ...  # type: ignore[override]



AsyncPipelineT = Union[AsyncPipeline, AsyncRetryablePipeline]
AsyncPubSubT = Union[AsyncPubSub, AsyncRetryablePubSub]

class AsyncKeyDB(AsyncRedis):
    """
    Implementation of the KeyDB protocol.

    This abstract class provides a Python interface to all KeyDB commands
    and an implementation of the KeyDB protocol.

    Pipelines derive from this, implementing how
    the commands are sent and received to the KeyDB server. Based on
    configuration, an instance will either use a AsyncConnectionPool, or
    AsyncConnection object to talk to redis.
    """

    @property
    def is_async(self):
        return True
    
    @classmethod
    def from_url(
        cls, 
        url: str, 
        pool_class: typing.Optional[typing.Type[AsyncConnectionPool]] = AsyncConnectionPool,
        single_connection_client: bool = False,
        **kwargs
    ):
        """
        Return a KeyDB client object configured from the given URL

        For example::
            keydb://[[username]:[password]]@localhost:6379/0
            keydbs://[[username]:[password]]@localhost:6379/0
            redis://[[username]:[password]]@localhost:6379/0
            rediss://[[username]:[password]]@localhost:6379/0
            unix://[[username]:[password]]@/path/to/socket.sock?db=0

        Five URL schemes are supported:
        - `keydb://` creates a TCP socket connection.
        - `keydbs://` creates a SSL wrapped TCP socket connection.
        - `redis://` creates a TCP socket connection. See more at:
          <https://www.iana.org/assignments/uri-schemes/prov/redis>
        - `rediss://` creates a SSL wrapped TCP socket connection. See more at:
          <https://www.iana.org/assignments/uri-schemes/prov/rediss>
        - ``unix://``: creates a Unix Domain Socket connection.

        The username, password, hostname, path and all querystring values
        are passed through urllib.parse.unquote in order to replace any
        percent-encoded values with their corresponding characters.

        There are several ways to specify a database number. The first value
        found will be used:
            1. A ``db`` querystring option, e.g. redis://localhost?db=0
            2. If using the redis:// or rediss:// schemes, the path argument
               of the url, e.g. redis://localhost/0
            3. A ``db`` keyword argument to this function.

        If none of these options are specified, the default db=0 is used.

        All querystring options are cast to their appropriate Python types.
        Boolean arguments can be specified with string values "True"/"False"
        or "Yes"/"No". Values that cannot be properly cast cause a
        ``ValueError`` to be raised. Once parsed, the querystring arguments
        and keyword arguments are passed to the ``AsyncConnectionPool``'s
        class initializer. In the case of conflicting arguments, querystring
        arguments always win.

        """
        if pool_class is None: pool_class = AsyncConnectionPool
        connection_pool = pool_class.from_url(url, **kwargs)
        return cls(
            connection_pool = connection_pool,
            single_connection_client = single_connection_client,
        )


    def pubsub(
        self, 
        retryable: typing.Optional[bool] = False,
        **kwargs,
    ) -> AsyncPubSubT:
        """
        Return a Publish/Subscribe object. With this object, you can
        subscribe to channels and listen for messages that get published to
        them.
        """
        pubsub_class = AsyncRetryablePubSub if retryable else AsyncPubSub
        return pubsub_class(connection_pool = self.connection_pool, **kwargs)
    

    def pipeline(
        self, 
        transaction: bool = True, 
        shard_hint: typing.Optional[str] = None,
        retryable: typing.Optional[bool] = False,
    ) -> AsyncPipelineT:
        """
        Return a new pipeline object that can queue multiple commands for
        later execution. ``transaction`` indicates whether all commands
        should be executed atomically. Apart from making a group of operations
        atomic, pipelines are useful for reducing the back-and-forth overhead
        between the client and server.
        """
        pipeline_class = AsyncRetryablePipeline if retryable else AsyncPipeline
        return pipeline_class(
            connection_pool = self.connection_pool, 
            response_callbacks = self.response_callbacks, 
            transaction = transaction, 
            shard_hint = shard_hint
        )




"""
Base Worker Class
"""
import os
import asyncio
import signal
import typing
import contextlib
import functools
from croniter import croniter
# from lazyops.utils.logs import default_logger as logger

from aiokeydb.v2.exceptions import ConnectionError
from aiokeydb.v2.configs import settings as default_settings
from aiokeydb.v2.types.jobs import Job, CronJob, JobStatus, TaskType
from aiokeydb.v2.utils.queue import (
    millis,
    now,
    seconds,
    get_hostname,
    get_and_log_exc,
    ensure_coroutine_function,
    get_thread_pool,
)
from aiokeydb.v2.utils.logs import logger, ColorMap

if typing.TYPE_CHECKING:
    from aiokeydb.v2.types.task_queue import TaskQueue
    from aiokeydb.v2.configs import KeyDBSettings


class Worker:
    """
    Worker is used to process and monitor jobs.

    queue: instance of saq.queue.Queue
    functions: list of async functions
    concurrency: number of jobs to process concurrently
    startup: async function to call on startup
    shutdown: async function to call on shutdown
    before_process: async function to call before a job processes
    after_process: async function to call after a job processes
    timers: dict with various timer overrides in seconds
        schedule: how often we poll to schedule jobs
        stats: how often to update stats
        sweep: how often to cleanup stuck jobs
        abort: how often to check if a job is aborted
    """

 

    SIGNALS = [signal.SIGINT, signal.SIGTERM] if os.name != "nt" else [signal.SIGTERM]

    def __init__(
        self,
        queue: 'TaskQueue',
        functions: typing.Optional[typing.List[typing.Callable]] = None,
        *,
        name: typing.Optional[str] = None,
        settings: typing.Optional['KeyDBSettings'] = None,
        concurrency: typing.Optional[int] = None,
        broadcast_concurrency: typing.Optional[int] = None,
        cron_jobs: typing.Optional[typing.List[typing.Callable]] = None,
        startup: typing.Optional[typing.Union[typing.List[typing.Callable], typing.Callable,]] = None,
        shutdown: typing.Optional[typing.Union[typing.List[typing.Callable], typing.Callable,]] = None,
        before_process: typing.Optional[typing.Callable] = None,
        after_process: typing.Optional[typing.Callable] = None,
        timers: typing.Optional[typing.Dict[str, typing.Union[int, float, typing.Any]]] = None,
        dequeue_timeout: int = 0,
        selectors: typing.Optional[typing.Dict[str, typing.Any]] = None,
        threadpool_size: typing.Optional[int] = None,
        debug_enabled: typing.Optional[bool] = None,
        silenced_functions: typing.Optional[typing.List[str]] = None,
        worker_attributes: typing.Optional[typing.Dict[str, typing.Any]] = None,
        heartbeat_ttl: typing.Optional[int] = None,
        is_leader_process: typing.Optional[bool] = None,
        verbose_startup: typing.Optional[bool] = None,
        verbose_concurrency: typing.Optional[bool] = True,
    ):  # sourcery skip: low-code-quality
        self.queue = queue
        self.settings = settings or default_settings
        self.worker_host = get_hostname()
        self.name = name if name is not None else (self.settings.worker.name or self.worker_host)
        self.concurrency = concurrency if concurrency is not None else self.settings.worker.concurrency
        self.broadcast_concurrency = (broadcast_concurrency if broadcast_concurrency is not None else self.settings.worker.max_broadcast_concurrency) or 3
        
        self.is_leader_process = is_leader_process if is_leader_process is not None else (self.settings.worker.is_leader_process if self.settings.worker.is_leader_process is not None else True)
        self.debug_enabled = debug_enabled if debug_enabled is not None else self.settings.worker.debug_enabled
        
        self.verbose_startup = verbose_startup if verbose_startup is not None else self.settings.worker.verbose_startup
        self.verbose_concurrency = verbose_concurrency

        if silenced_functions:
            self.queue.add_silenced_functions(*silenced_functions)

        self.startup = startup or [
            self.settings.worker.get_startup_context
        ]
        if not isinstance(self.startup, list):
            self.startup = [self.startup]
        self.shutdown = shutdown or [
            self.settings.worker.run_shutdown_funcs
        ]
        if not isinstance(self.shutdown, list):
            self.shutdown = [self.shutdown]
        
        self.before_process = before_process
        self.after_process = after_process
        self.timers = {
            "schedule": 1,
            "stats": 60,
            "sweep": 180,
            "abort": 1,
            "heartbeat": 5,
            "broadcast": 10,
        }
        timers = timers or self.settings.worker.timers
        if timers: self.timers.update(timers)
        self.selectors = selectors or {}
        self.event = asyncio.Event()
        functions: typing.Set[typing.Callable] = set(functions or self.settings.worker.get_functions(verbose = self.debug_enabled))
        self.functions = {}
        self.cron_jobs: typing.List[CronJob] = cron_jobs or self.settings.worker.get_cronjobs()
        self.tasks: typing.Set[asyncio.Task] = set()
        self.job_task_contexts: typing.Dict['Job', typing.Dict[str, typing.Any]] = {}
        self.dequeue_timeout = dequeue_timeout
        self.worker_pid: int = os.getpid()
        self.worker_attributes = worker_attributes or {}
        self.worker_attributes.update({
            # "name": self.name,
            "selectors": self.selectors,
            "worker_host": self.worker_host,
            "worker_id": self.queue.uuid,
            "queue_name": self.queue.queue_name,
            "is_leader_process": self.is_leader_process,
            "worker_pid": self.worker_pid,
        })
        self.heartbeat_ttl = heartbeat_ttl if heartbeat_ttl is not None else self.queue.heartbeat_ttl

        for job in self.cron_jobs:
            if not croniter.is_valid(job.cron):
                raise ValueError(f"Cron is invalid {job.cron}")
            self.functions[job.function_name] = job.function

        for function in functions:
            if isinstance(function, tuple): name, function = function
            else: name = function.__qualname__
            self.functions[name] = function
        self._worker_identity: str = f'[{self.worker_pid}] {self.settings.app_name or "KeyDBWorker"}'
    
    @property
    def _worker_name(self):
        return f"{self.worker_host}.{self.name}.{self.worker_pid}"
    
    @property
    def _is_ctx_retryable(self) -> bool:
        """
        Returns whether the context is retryable.
        """
        return hasattr(self.queue.ctx.async_client, "_is_retryable_wrapped")
    
    @property
    def _is_primary_worker(self) -> bool:
        """
        Returns whether the worker is the primary worker.
        """
        return self.is_leader_process or self.name[-1] == '0'


    def is_silenced_function(
        self,
        name: str,
        stage: typing.Optional[str] = None,
    ) -> bool:
        """
        Checks if a function is silenced
        """
        return self.queue.is_silenced_function(name, stage = stage)


    def logger(self, job: 'Job' = None, kind: str = "enqueue"):
        """
        The logger for the worker.
        """
        if job:
            return logger.bind(
                worker_name = self._worker_name,
                job_id = job.id,
                status = job.status,
                queue_name = getattr(job.queue, 'queue_name', self.queue_name) or 'unknown queue',
                kind = kind,
            )
        else:
            return logger.bind(
                worker_name = self._worker_name,
                queue_name = self.queue_name,
                kind = kind,
            )

    async def _before_process(self, ctx):
        """
        Handles the before process function.
        """
        if self.before_process:
            await self.before_process(ctx)

    async def _after_process(self, ctx):
        """
        Handles the after process function.
        """
        if self.after_process:
            await self.after_process(ctx)
    
    async def prepare_server(self):
        """
        Prepares the keydb server to ensure that the maximum number of concurrent
        connections are available.
        """
        await self.queue.prepare_server()

    def _get_startup_log_message(self):  # sourcery skip: low-code-quality
        """
        Builds the startup log message.
        """
        _msg = f'{self._worker_identity}: {self.worker_host}.{self.name} v{self.settings.version}'
        _msg += f'\n- {ColorMap.cyan}[Worker ID]{ColorMap.reset}: {ColorMap.bold}{self.worker_id}{ColorMap.reset}'
        if self._is_primary_worker:
            _msg += f'\n- {ColorMap.cyan}[Queue]{ColorMap.reset}: {ColorMap.bold}{self.queue_name} @ {self.queue.uri} DB: {self.queue.db_id}{ColorMap.reset}'
            _msg += f'\n- {ColorMap.cyan}[Registered]{ColorMap.reset}: {ColorMap.bold}{len(self.functions)} functions, {len(self.cron_jobs)} cron jobs{ColorMap.reset}'
            _msg += f'\n- {ColorMap.cyan}[Concurrency]{ColorMap.reset}: {ColorMap.bold}{self.concurrency}/jobs, {self.broadcast_concurrency}/broadcasts{ColorMap.reset}'
            if self.verbose_startup:

                _msg += f'\n- {ColorMap.cyan}[Serializer]{ColorMap.reset}: {self.queue.serializer}'
                _msg += f'\n- {ColorMap.cyan}[Worker Attributes]{ColorMap.reset}: {self.worker_attributes}'
                if self._is_ctx_retryable:
                    _msg += f'\n- {ColorMap.cyan}[Retryable]{ColorMap.reset}: {self._is_ctx_retryable}'
                _msg += f'\n- {ColorMap.cyan}[Functions]{ColorMap.reset}:'
                for function_name in self.functions:
                    _msg += f'\n   - {ColorMap.bold}{function_name}{ColorMap.reset}'
                if self.settings.worker.has_silenced_functions:
                    _msg += f"\n - {ColorMap.cyan}[Silenced Functions]{ColorMap.reset}:"
                    for stage, silenced_functions in self.settings.worker.silenced_function_dict.items():
                        if silenced_functions:
                            _msg += f"\n   - {stage}: {silenced_functions}"
                if self.queue.function_tracker_enabled:
                    _msg += f'\n- {ColorMap.cyan}[Function Tracker Enabled]{ColorMap.reset}: {self.queue.function_tracker_enabled}'
        return _msg
            

    async def start(self):
        """
        Start processing jobs and upkeep tasks.
        """
        os.environ['IS_WORKER_PROCESS'] = 'true'
        if not self.settings.worker_enabled:
            self.logger(kind = "startup").warning(
                f'{self._worker_identity}: {self.worker_host}.{self.name} is disabled.')
            return
        
        self.context = {"worker": self, "queue": self.queue, "keydb": self.queue.ctx, "pool": get_thread_pool(), "vars": {}}
        self.worker_attributes['name'] = self.name
        self.queue._worker_name = self._worker_name

        try:
            self.event = asyncio.Event()
            loop = asyncio.get_running_loop()

            for signum in self.SIGNALS: loop.add_signal_handler(signum, self.event.set)
            await self.queue.ctx.async_client.initialize()

            if self.startup: 
                for func in self.startup:
                    await func(self.context)
            
            self.logger(kind = "startup").info(self._get_startup_log_message())
            # if self.queue.function_tracker_enabled:
            #     await self.queue.setup_function_tracker(self.functions)

            # Register the queue
            # await self.queue.register_queue()
            # Send the first heartbeat
            await self.heartbeat()

            # Prepare the server
            await self.prepare_server()

            self.tasks.update(await self.upkeep())
            for cid in range(self.concurrency):
                self._process(concurrency_id = cid)
            
            for _ in range(self.broadcast_concurrency):
                self._broadcast_process()

            await self.event.wait()
        finally:
            self.logger(kind = "shutdown").warning(f'{self._worker_identity}: {self.worker_host}.{self.name} is shutting down.')
            if self.shutdown:
                for func in self.shutdown:
                    await func(self.context)
            await self.stop()
    

    async def stop(self):
        """
        Stop the worker and cleanup.
        """
        self.event.set()
        all_tasks = list(self.tasks)
        self.tasks.clear()
        for task in all_tasks:
            task.cancel()
        await asyncio.gather(*all_tasks, return_exceptions=True)
    

    async def schedule(self, lock: int = 1):
        """
        Schedule jobs.
        """
        for cron_job in self.cron_jobs:
            enqueue_kwargs = cron_job.to_enqueue_kwargs(
                job_key = self.queue.job_id(f"cron:{cron_job.function_name}") if cron_job.unique else None,
                exclude_none = True,
            )
            await self.queue.enqueue(**enqueue_kwargs)
            

        scheduled = await self.queue.schedule(lock)
        if scheduled and self.queue.verbose_results:
            self.logger(kind = "scheduled").info(
                f'↻ node={self.queue.node_name}, {scheduled}'
            )

    async def heartbeat(self, ttl: int = 20):
        """
        Send a heartbeat to the queue.
        """
        # self.logger(kind = "scheduled").info(f"❤ Sending [{self.worker_id}] heartbeat to queue {self.queue_name}")
        await self.queue.add_heartbeat(
            worker_id = self.worker_id,
            worker_attributes = self.worker_attributes,
            heartbeat_ttl = ttl,
        )

    async def upkeep(self):
        """
        Start various upkeep tasks async.
        """
        async def poll(func, sleep, arg=None, **kwargs):
            while not self.event.is_set():
                try:
                    if asyncio.iscoroutinefunction(func):
                        await func(arg or sleep, **kwargs)
                    else: func(arg or sleep, **kwargs)
                except (Exception, asyncio.CancelledError):
                    if self.event.is_set(): return
                    get_and_log_exc(func = func)
                await asyncio.sleep(sleep)

        return [
            asyncio.create_task(poll(self.abort, self.timers["abort"])),
            asyncio.create_task(poll(self.schedule, self.timers["schedule"])),
            asyncio.create_task(poll(self.queue.sweep, self.timers["sweep"])),
            asyncio.create_task(
                poll(self.queue.stats, self.timers["stats"], self.timers["stats"] + 1)
            ),
            asyncio.create_task(
                poll(self.heartbeat, self.timers["heartbeat"], self.heartbeat_ttl)
            ),

        ]

    async def abort(self, abort_threshold: int):
        """
        Abort jobs that have been running for too long.
        """
        jobs = [
            job
            for job in self.job_task_contexts
            if job.duration("running") >= millis(abort_threshold)
        ]

        if not jobs: return
        aborted = await self.queue.ctx.async_mget(job.abort_id for job in jobs)
        for job, abort in zip(jobs, aborted):
            if not abort: continue

            task_data = self.job_task_contexts.get(job, {})
            task: asyncio.Task = task_data.get("task")

            if task and not task.done():
                task_data["aborted"] = True
                task.cancel()
                await asyncio.gather(task, return_exceptions=True)

            await job.finish(JobStatus.ABORTED, error = abort.decode("utf-8"))
            await self.queue.ctx.async_delete(job.abort_id)
            if not self.is_silenced_function(job.function, stage = "abort"):
                self.logger(job = job, kind = "abort").info(f"⊘ {job.duration('running')}ms, node={self.node_name}, func={job.function}, id={job.id}")
    
    async def process(self, broadcast: typing.Optional[bool] = False, concurrency_id: typing.Optional[int] = None):
        """
        Process a job.
        """
        # sourcery skip: low-code-quality
        # pylint: disable=too-many-branches
        job, context = None, None
        try:
            with contextlib.suppress(ConnectionError):
                job = await self.queue.dequeue(
                    self.dequeue_timeout, 
                    worker_id = self.worker_id if broadcast else None, 
                    worker_name = self.name if broadcast else None,
                )

            if not job: return
            if job.worker_id and job.worker_id != self.worker_id:
                if self.debug_enabled or self.queue.verbose_results:
                    self.logger(job = job, kind = "process").info(f"⊘ Rejected job, queued_key={job.queued_key}, func={job.function}, id={job.id} | worker_id={job.worker_id} != {self.worker_id}")
                return
            if job.worker_name and job.worker_name != self.name:
                if self.debug_enabled or self.queue.verbose_results:
                    self.logger(job = job, kind = "process").info(f"⊘ Rejected job, queued_key={job.queued_key}, func={job.function}, id={job.id} | worker_name={job.worker_name} != {self.name}")
                return
                
            if job.worker_name or job.worker_id and (self.debug_enabled or self.queue.verbose_results):
                self.logger(job = job, kind = "process").info(f"☑ Accepted job, func={job.function}, id={job.id}, worker_name={job.worker_name}, worker_id={job.worker_id}")

            job.started = now()
            job.status = JobStatus.ACTIVE
            job.attempts += 1
            await job.update()
            if self.queue.function_tracker_enabled:
                await self.queue.track_job_id(job)
            context = {**self.context, "job": job}
            await self._before_process(context)
            # if job.function not in self.silenced_functions:
            if not self.is_silenced_function(job.function, stage = "process"):
                _msg = f"← duration={job.duration('running')}ms, node={self.node_name}, func={job.function}"
                if self.verbose_concurrency:
                    _msg = _msg.replace("node=", f"conn=[{concurrency_id}/{self.concurrency}], node=")
                self.logger(job = job, kind = "process").info(_msg)

            function = ensure_coroutine_function(self.functions[job.function])
            task = asyncio.create_task(function(context, **(job.kwargs or {})))
            self.job_task_contexts[job] = {"task": task, "aborted": False}
            result = await asyncio.wait_for(task, job.timeout)
            await job.finish(JobStatus.COMPLETE, result=result)
        except asyncio.CancelledError:
            if job and not self.job_task_contexts.get(job, {}).get("aborted"):
                await job.retry("cancelled")
        except Exception:
            error = get_and_log_exc(job = job)

            if job:
                if job.attempts >= job.retries:
                    await job.finish(JobStatus.FAILED, error=error)
                else: await job.retry(error)
        finally:
            if context:
                self.job_task_contexts.pop(job, None)
                try: await self._after_process(context)
                except (Exception, asyncio.CancelledError): 
                    get_and_log_exc(job = job)

    async def process_broadcast(self):
        """
        This is a separate process that runs in the background to process broadcasts.
        """
        # await self.heartbeat(self.heartbeat_ttl)
        await self.process(broadcast = True)

        await self.queue.schedule(lock = 1, worker_id = self.worker_id)
        await self.queue.schedule(lock = 1, worker_id = self.name)

        await self.queue.sweep(worker_id = self.worker_id)
        await self.queue.sweep(worker_id = self.name)


    def _process(self, previous_task=None, concurrency_id: typing.Optional[int] = None):
        """
        Handles the processing of jobs.
        """
        if previous_task: self.tasks.discard(previous_task)
        if not self.event.is_set():
            new_task = asyncio.create_task(self.process(concurrency_id = concurrency_id))
            self.tasks.add(new_task)
            if concurrency_id is not None:
                new_task.add_done_callback(functools.partial(self._process, concurrency_id = concurrency_id))
            else:
                new_task.add_done_callback(self._process)
    
    def _broadcast_process(self, previous_task = None):
        """
        This is a separate process that runs in the background to process broadcasts.
        """
        if previous_task and isinstance(previous_task, asyncio.Task):
            self.tasks.discard(previous_task)

        if not self.event.is_set():
            new_task = asyncio.create_task(self.process_broadcast())
            self.tasks.add(new_task)
            new_task.add_done_callback(self._broadcast_process)
        
    
    @classmethod
    def from_settings(
        cls,
        settings: typing.Union[str, typing.Dict],
        queue: typing.Optional['TaskQueue'] = None,
        verbose: typing.Optional[bool] = None,
        **kwargs
    ) -> "Worker":
        """
        Create a Worker from a settings file or dict.
        """
        verbose = verbose or default_settings.worker.debug_enabled
        if isinstance(settings, str):
            import importlib
            if verbose: logger.info(f"Importing settings from {settings}")
            module_path, name = settings.strip().rsplit(".", 1)
            module = importlib.import_module(module_path)
            settings = getattr(module, name)
        elif verbose: logger.info(f"Using settings dict {settings}")
        if kwargs: settings.update(kwargs)
        if queue is not None: settings['queue'] = queue
        return cls(**settings)


    """
    Properties
    """
    @property
    def worker_id(self) -> str:
        """
        Returns a unique worker id.
        """
        return self.queue.uuid

    @property
    def queue_name(self) -> str:
        """
        Returns the queue name.
        """
        return self.queue.queue_name

    @property
    def node_name(self) -> str:
        """
        Returns the node name.
        """
        return self.queue.node_name
    

    """
    Static Methods to add functions and context to the worker queue.
    """

    @staticmethod
    def add_context(
        obj: typing.Optional[typing.Any] = None,
        name: typing.Optional[str] = None,
        verbose: typing.Optional[bool] = None,
        silenced: typing.Optional[bool] = None,
        _fx: typing.Optional[typing.Callable] = None,
        **kwargs,
    ):
        """
        Add a context object to the worker queue.

        Args:
            name: name of the job
            obj: object to pass to the function
            verbose: whether to print the function's output
            kwargs: additional arguments to pass to the function

        >> @Worker.add_context("client")
        >> async def init_client():
        >>     return Client()

        >> ctx['client'] = Client
            
        """
        return default_settings.worker.add_context(
            obj = obj, 
            name = name, 
            verbose = verbose, 
            silenced = silenced, 
            _fx = _fx, 
            **kwargs
        )

    @staticmethod
    def add_dependency(
        obj: typing.Optional[typing.Any] = None,
        name: typing.Optional[str] = None,
        verbose: typing.Optional[bool] = None,
        _fx: typing.Optional[typing.Callable] = None,
        silenced: typing.Optional[bool] = None,
        **kwargs,
    ):
        """
        Adds a dependency function to the worker queue.
        Args:
            name: name of the job
            obj: object to pass to the function
            verbose: whether to print the function's output
            kwargs: additional arguments to pass to the function
        """
        return default_settings.worker.add_dependency(
            obj = obj, 
            name = name, 
            verbose = verbose, 
            silenced = silenced, 
            _fx = _fx, 
            **kwargs
        )


    @staticmethod
    def on_startup(
        name: typing.Optional[str] = None,
        verbose: typing.Optional[bool] = None,
        _fx: typing.Optional[typing.Callable] = None,
        silenced: typing.Optional[bool] = None,
        **kwargs,
    ):
        """
        Add a startup function to the worker queue.


        >> @Worker.on_startup("client")
        >> async def init_client():
        >>     return Client()

        >> ctx['client'] = Client


        >> @Worker.on_startup(_set_ctx = True)
        >> async def init_client(ctx: Dict[str, Any], **kwargs):
        >>   ctx['client'] = Client()
        >>   return ctx

        >> ctx['client'] = Client

        
        """
        return default_settings.worker.on_startup(
            name = name, 
            verbose = verbose, 
            _fx = _fx, 
            silenced = silenced, 
            **kwargs
        )

    @staticmethod
    def on_shutdown(
        name: typing.Optional[str] = None,
        verbose: typing.Optional[bool] = None,
        _fx: typing.Optional[typing.Callable] = None,
        silenced: typing.Optional[bool] = None,
        **kwargs,
    ):
        """
        Add a shutdown function to the worker queue.
        """
        return default_settings.worker.on_shutdown(
            name = name, 
            verbose = verbose, 
            _fx = _fx, 
            silenced = silenced, 
            **kwargs
        )

    @staticmethod
    def add_function(
        name: typing.Optional[str] = None,
        _fx: typing.Optional[typing.Callable] = None,
        verbose: typing.Optional[bool] = None,
        silenced: typing.Optional[bool] = None,
        silenced_stages: typing.Optional[typing.List[str]] = None,
        **kwargs,
    ):
        """
        Add a function to the worker queue.

        >> @Worker.add_function()
        >> async def my_function(*args, **kwargs):
        >>     return "Hello World"

        >> res = await queue.apply('my_function', *args, **kwargs)
        >> assert res == "Hello World"
        
        """
        return default_settings.worker.add_function(
            name = name, 
            _fx = _fx, 
            verbose = verbose, 
            silenced = silenced, 
            silenced_stages = silenced_stages, 
            **kwargs
        )
    
    @staticmethod
    def add_function_to_silenced(
        name: str,
        silenced_stages: typing.Optional[typing.List[str]] = None,
        **kwargs
    ):
        """
        Add a function to the silenced functions list.
        """
        default_settings.worker.add_function_to_silenced(name = name, silenced_stages = silenced_stages, **kwargs)
    

    @staticmethod
    def add_cronjob(        
        schedule: typing.Optional[typing.Union[typing.Dict, typing.List, str]] = None,
        _fx: typing.Optional[typing.Callable] = None,
        verbose: typing.Optional[bool] = None,
        silenced: typing.Optional[bool] = None,
        silenced_stages: typing.Optional[typing.List[str]] = None,
        callback: typing.Optional[typing.Union[str, typing.Callable]] = None,
        callback_kwargs: typing.Optional[dict] = None,
        **kwargs,
    ):
        """
        Adds a function to `WorkerTask.cronjobs`.
        WorkerCronFuncs = {
            {'coroutine': refresh_spot_data, 'name': 'refresh_spot_data', 'minute': {10, 30, 50}},
        }
        """
        return default_settings.worker.add_cronjob(
            schedule = schedule, 
            _fx = _fx, 
            verbose = verbose, 
            silenced = silenced, 
            silenced_stages = silenced_stages, 
            callback = callback,
            callback_kwargs = callback_kwargs,
            **kwargs
        )
    
    
    @staticmethod
    def add(
        task: TaskType = TaskType.default,
        name: typing.Optional[str] = None,
        obj: typing.Optional[typing.Any] = None,
        verbose: typing.Optional[bool] = None,
        **kwargs,
    ):
        """
        Add a function to the worker queue.

        Args:
            func: function to call
            name: name of the job
            obj: object to pass to the function
            verbose: whether to print the function's output
            kwargs: additional arguments to pass to the function
        """
        return default_settings.worker.add(
            task = task, 
            name = name, 
            obj = obj, 
            verbose = verbose, 
            **kwargs
        )
    
    @staticmethod
    def set_queue_func(
        queue_func: typing.Union[typing.Callable, 'TaskQueue'],
    ):
        """
        Set the queue function to use for the worker queue.
        """
        return default_settings.worker.set_queue_func(queue_func = queue_func)

    @staticmethod
    def add_fallback_function(
        verbose: typing.Optional[bool] = True,
        silenced: typing.Optional[bool] = None,
        method: typing.Optional[str] = "apply",
        timeout: typing.Optional[int] = None,
        suppressed_exceptions: typing.Optional[typing.List[typing.Type[Exception]]] = None,
        failed_results: typing.Optional[typing.List] = None,
        queue_func: typing.Optional[typing.Union[typing.Callable, 'TaskQueue']] = None,
        silenced_stages: typing.Optional[typing.List[str]] = None,
        **kwargs,
    ):
        """
        Add a fallback function to the worker queue.
        The function will first be attempted in the worker queue
        and if it fails, it will be called directly.
        This is designed to reduce redundancy in writing fallback functions.
        
        **IMPORTANT**: The function must be a coroutine function, and
        and should only have kwargs

        >> @Worker.add_fallback_function()
        >> async def run_func(ctx: Optional[dict] = None, **kwargs):
        >>     if ctx:
        >>         print("Calling from Worker Queue. Will Fail.")
        >>         raise Exception("Failed")
        >>     print("Calling Directly")
        >>     return "Hello World"
        """
        return default_settings.worker.add_fallback_function(
            verbose = verbose, 
            silenced = silenced, 
            method = method, 
            timeout = timeout,
            suppressed_exceptions = suppressed_exceptions, 
            failed_results = failed_results,
            queue_func = queue_func, 
            silenced_stages = silenced_stages,
            **kwargs
        )


from __future__ import annotations

import os
import time
import uuid
import random
import typing
import logging
import traceback
import contextlib
import contextvars
import asyncio
import functools
import inspect

from concurrent import futures
from lazyops.utils.helpers import import_function, timer, build_batches
from lazyops.utils.ahelpers import amap_v2 as concurrent_map

logger = logging.getLogger(__name__)

_NodeName: str = None
_ThreadPool: futures.ThreadPoolExecutor = None

if typing.TYPE_CHECKING:
    from aiokeydb.v2.types.jobs import Job

def get_hostname() -> str:
    """
    Lazily initialize the worker node name
    """
    global _NodeName
    if _NodeName is None:
        if os.getenv('HOSTNAME'):
            _NodeName = os.getenv('HOSTNAME')
        else:
            import socket
            while _NodeName is None:
                with contextlib.suppress(Exception):
                    _NodeName = socket.gethostname()
                time.sleep(0.5)
    return _NodeName


def get_thread_pool(
    n_workers: int = None
) -> futures.ThreadPoolExecutor:
    """
    Lazily initialize the worker thread pool
    """
    global _ThreadPool
    if _ThreadPool is None:
        if n_workers is None:
            from aiokeydb.v2.configs import settings
            n_workers = settings.worker.threadpool_size
        _ThreadPool = futures.ThreadPoolExecutor(max_workers = n_workers)
    return _ThreadPool


async def run_in_executor(ctx: typing.Dict[str, typing.Any], func: typing.Callable, *args, **kwargs):
    blocking = functools.partial(func, *args, **kwargs)
    loop = asyncio.get_running_loop()
    return await loop.run_in_executor(ctx['pool'], blocking)


def get_func_name(func: typing.Union[str, typing.Callable]) -> str:
    """
    Returns the function name
    """
    return func.__name__ if callable(func) else func


def get_func_full_name(func: typing.Union[str, typing.Callable]) -> str:
    """
    Returns the function name
    """
    return f"{func.__module__}.{func.__qualname__}" if callable(func) else func

def get_and_log_exc(
    job: typing.Optional['Job'] = None, 
    func: typing.Optional[typing.Union[str, typing.Callable]] = None
):
    error = traceback.format_exc()
    err_msg = f'node={get_hostname()}, {error}'
    if func: err_msg = f'func={get_func_name(func)}, {err_msg}'
    elif job: err_msg = f'job={job.short_repr}, {err_msg}'
    logger.error(err_msg)
    return error



def now():
    return int(time.time() * 1000)

def uuid1():
    return str(uuid.uuid1())

def uuid4():
    return str(uuid.uuid4())

def millis(s):
    return s * 1000

def seconds(ms):
    return ms / 1000


def exponential_backoff(
    attempts,
    base_delay,
    max_delay=None,
    jitter=True,
):
    """
    Get the next delay for retries in exponential backoff.

    attempts: Number of attempts so far
    base_delay: Base delay, in seconds
    max_delay: Max delay, in seconds. If None (default), there is no max.
    jitter: If True, add a random jitter to the delay
    """
    if max_delay is None:
        max_delay = float("inf")
    backoff = min(max_delay, base_delay * 2 ** max(attempts - 1, 0))
    if jitter:
        backoff = backoff * random.random()
    return backoff


_JobKeyMethod = {
    'uuid1': uuid1,
    'uuid4': uuid4,
}

_JobKeyFunc: typing.Callable = None

def _get_jobkey_func():
    global _JobKeyFunc
    if _JobKeyFunc is None:
        from aiokeydb.v2.configs import settings
        _JobKeyFunc = _JobKeyMethod[settings.worker.job_key_method]
    return _JobKeyFunc()

def get_default_job_key():
    return _get_jobkey_func()

def isclassmethod(method):
    bound_to = getattr(method, '__self__', None)
    if not isinstance(bound_to, type):
        # must be bound to a class
        return False
    name = method.__name__
    for cls in bound_to.__mro__:
        descriptor = vars(cls).get(name)
        if descriptor is not None:
            return isinstance(descriptor, classmethod)
    return False

def ensure_coroutine_function(func):
    if asyncio.iscoroutinefunction(func):
        return func

    async def wrapped(*args, **kwargs):
        loop = asyncio.get_running_loop()
        ctx = contextvars.copy_context()
        return await loop.run_in_executor(
            executor = None, func = lambda: ctx.run(func, *args, **kwargs)
        )

    return wrapped


_bg_tasks: typing.Set[asyncio.Task] = set()


def create_background_task(func: typing.Union[typing.Callable, typing.Coroutine], *args, **kwargs):
    """
    Creates a background task and adds it to the global set of background tasks
    """
    if inspect.isawaitable(func):
        task = asyncio.create_task(func)
    else:
        task = asyncio.create_task(func(*args, **kwargs))
    _bg_tasks.add(task)
    task.add_done_callback(_bg_tasks.discard)
    return task



from lazyops.utils.logs import get_logger, STATUS_COLOR, COLORED_MESSAGE_MAP, FALLBACK_STATUS_COLOR


class ColorMap:
    green: str = '\033[0;32m'
    red: str = '\033[0;31m'
    yellow: str = '\033[0;33m'
    blue: str = '\033[0;34m'
    magenta: str = '\033[0;35m'
    cyan: str = '\033[0;36m'
    white: str = '\033[0;37m'
    bold: str = '\033[1m'
    reset: str = '\033[0m'




class CustomizeLogger:

    @staticmethod
    def logger_formatter(record: dict) -> str:
        """
        To add a custom format for a module, add another `elif` clause with code to determine `extra` and `level`.

        From that module and all submodules, call logger with `logger.bind(foo='bar').info(msg)`.
        Then you can access it with `record['extra'].get('foo')`.
        """        
        extra = '<cyan>{name}</>:<cyan>{function}</>: '

        if record.get('extra'):
            if record['extra'].get('request_id'):
                extra = '<cyan>{name}</>:<cyan>{function}</>:<green>request_id: {extra[request_id]}</> '

            elif record['extra'].get('job_id') and record['extra'].get('queue_name') and record['extra'].get('kind'):
                status = record['extra'].get('status')
                color = STATUS_COLOR.get(status, FALLBACK_STATUS_COLOR)
                kind_color = STATUS_COLOR.get(record.get('extra', {}).get('kind'), FALLBACK_STATUS_COLOR)
                if not record['extra'].get('worker_name'):
                    record['extra']['worker_name'] = ''
                extra = '<cyan>{extra[queue_name]}</>:<bold><magenta>{extra[worker_name]}</></>:<bold><' + kind_color + '>{extra[kind]:<9}</></> <' + color + '>{extra[job_id]}</> '
                # if record['extra'].get('cid'):
                #     extra += '<green>({extra[cid]}/{extra[conn]})</> '

            elif record['extra'].get('kind') and record['extra'].get('queue_name'):
                if not record['extra'].get('worker_name'):
                    record['extra']['worker_name'] = ''
                kind_color = STATUS_COLOR.get(record.get('extra', {}).get('kind'), FALLBACK_STATUS_COLOR)
                extra = '<cyan>{extra[queue_name]}</>:<b><magenta>{extra[worker_name]}</></>:<b><' + kind_color + '>{extra[kind]:<9}</></> '
                # if record['extra'].get('cid'):
                #     extra += '<green>({extra[cid]}/{extra[conn]})</> '


        if 'result=tensor([' not in str(record['message']):
            return "<level>{level: <8}</> <green>{time:YYYY-MM-DD HH:mm:ss.SSS}</>: "\
                       + extra + "<level>{message}</level>\n"
        msg = str(record['message'])[:100].replace('{', '(').replace('}', ')')
        return "<level>{level: <8}</> <green>{time:YYYY-MM-DD HH:mm:ss.SSS}</>: "\
                   + extra + "<level>" + msg + f"</level>{STATUS_COLOR['reset']}\n"


logger = get_logger(
    format = CustomizeLogger.logger_formatter,
)
